import json

import boto3
import os
import logging

from .db_util import format_sql, render_data_api_response


class RdsDataDao:

    def __init__(self, db, cluster_arn=None, secret_arn=None, **kwargs):
        self.db = db
        self.cluster_arn = cluster_arn or os.getenv('DB_URL')
        self.secret_arn = secret_arn or os.getenv('DB_CREDENTIALS')
        self.logger = logging.getLogger(kwargs.get('logger_name', db))
        endpoint_url = os.getenv('DB_ENDPOINT_URL', None)
        if endpoint_url:
            self.rds_data = boto3.client('rds-data', endpoint_url=endpoint_url)
        else:
            self.rds_data = boto3.client('rds-data')

    def __str__(self):
        return '{}: {} {}'.format(self.__class__.__name__, self.db, self.cluster_arn)

    def insert(self, cmd, data=None, multiple_returns=False):
        items = self._execute(cmd, True, data=data)
        return self.get_first(items) if not multiple_returns else items

    def get(self, cmd, data=None):
        return self._execute(cmd, data=data)

    def get_objects(self, cmd, parsing_class, data=None):
        return [parsing_class.parse_obj(item) for item in self._execute(cmd, data=data)]

    def delete(self, cmd, data=None):
        self._execute(cmd, False, data)

    def get_single_result(self, cmd, data=None):
        result = self._execute(cmd, data=data)
        return self.get_first(result)

    def get_single_result_object(self, cmd, parsing_class, data=None):
        result = self._execute(cmd, data=data)
        single_result = self.get_first(result)
        if single_result:
            return parsing_class.parse_obj(single_result)
        return None

    def update(self, cmd, data=None):
        return self._execute(cmd, True, data=data)

    def transaction_begin(self):
        response = self.rds_data.begin_transaction(resourceArn=self.cluster_arn,
                                                   secretArn=self.secret_arn,
                                                   database=self.db)
        if 'transactionId' not in response:
            raise Exception('No transaction id found in response', response)
        transaction_id = response['transactionId']
        self.logger.debug('tx ' + str(transaction_id))
        return transaction_id

    def get_first(self, items):
        return items[0] if (items and len(items) > 0) else None

    def transaction_insert(self, transaction_id, cmd, data=None):
        self.logger.debug('tx ' + str(transaction_id))
        try:
            items = self._execute(cmd, True, data=data, transaction_id=transaction_id)
            return self.get_first(items)
        except Exception as e:
            print(f'Error: {e}')
            self.rds_data.rollback_transaction(resourceArn=self.cluster_arn,
                                               secretArn=self.secret_arn, transactionId=transaction_id)
            raise e

    def transaction_end(self, transaction_id):
        self.logger.debug('tx ' + str(transaction_id))
        try:
            self.rds_data.commit_transaction(resourceArn=self.cluster_arn,
                                             secretArn=self.secret_arn, transactionId=transaction_id)
        except Exception as e:
            print(f'Error: {e}')
            self.rds_data.rollback_transaction(resourceArn=self.cluster_arn,
                                               secretArn=self.secret_arn, transactionId=transaction_id)
            raise e

    def _execute(self, cmd, insert=False, data=None, transaction_id=''):
        if not cmd:
            return
        will_return = not insert or 'returning' in cmd.lower()
        sql, parameters = format_sql(cmd, data)
        self.logger.debug('query: %s' % {'sql': sql, 'parameters': json.dumps(parameters)})
        try:
            # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds-data.html#RDSDataService.Client.execute_statement
            response = self.rds_data.execute_statement(
                resourceArn=self.cluster_arn,
                secretArn=self.secret_arn,
                database=self.db,
                includeResultMetadata=True,
                resultSetOptions={
                    'decimalReturnType': 'STRING'
                },
                transactionId=transaction_id,
                parameters=parameters,
                sql=sql)
            return render_data_api_response(response).get('records', None) if will_return else None
        except Exception as e:
            self.logger.error('error executing command: %s' % {'cmd': cmd}, exc_info=e)
            raise e



from setuptools import setup

setup(name='rds_data_dao',
      version='0.1.10',
      description='RDS Data API Wrapper',
      url='https://github.com/stavvy/rds-data-dao',
      author='Chris Buonocore',
      author_email='chris@stavvy.com',
      license='MIT',
      test_suite='tests',
      setup_requires=['pytest-runner'],
      tests_require=['pytest'],
      packages=['rds_data_dao'],
      zip_safe=False)


import itertools as it
import json
import re
from datetime import datetime
from urllib import parse


def safe_escape_string(s):
    # Filter nonstandard characters from input, need to fix for complex object/json inputs.
    return re.sub('[^-|.:_A-Za-z0-9 ]+', '', str(s))




def _render_value(value):
    if value.get("isNull"):
        return None
    elif "arrayValue" in value:
        if "arrayValues" in value["arrayValue"]:
            return [_render_value(nested) for nested in value["arrayValue"]["arrayValues"]]
        else:
            return list(value["arrayValue"].values())[0]
    else:
        return list(value.values())[0]


def parse_db_timestamp(d):
    if type(d) == str:
        d = datetime.strptime(d, '%Y-%m-%d %H:%M:%S.%f')
    return datetime.timestamp(d)


def render_data_api_response(response):
    if "records" in response:
        column_names = list(map(lambda x: x['name'], response.get('columnMetadata', [])))
        print(column_names, response['records'])
        for i, record in enumerate(response["records"]):
            response["records"][i] = {column_names[j]: _render_value(v) for j, v in enumerate(response["records"][i])}
    return response


DATA_API_TYPE_MAP = {
    bytes: "blobValue",
    dict: "stringValue",
    bool: "booleanValue",
    float: "doubleValue",
    int: "longValue",
    str: "stringValue",
    list: "arrayValue",
    tuple: "arrayValue"
}


def get_data_type(x):
    return DATA_API_TYPE_MAP.get(type(x), 'stringValue')


def is_json(myjson):
    if type(myjson) == 'str':
        try:
            json_object = json.loads(myjson)
        except ValueError as e:
            return False
    return True


# ex: x::org_types -> org_types
def get_cast(format_match):
    return format_match[2:]


# ref: https://github.com/chanzuckerberg/aurora-data-api
# https://aws.amazon.com/blogs/database/using-the-data-api-to-interact-with-an-amazon-aurora-serverless-mysql-database/
# https://forums.aws.amazon.com/thread.jspa?messageID=921843
def format_sql(raw_sql, data):
    # using parameterized str for now.
    if not data:
        data = []
    data = tuple(map(stringify_for_data_api_query, data))

    def parameterize_string(s):
        cnt = it.count()
        return re.sub(r"(%s[a-zA-Z_:]*)", lambda x: ':name{}{}'.format(next(cnt), get_cast(x.group(0))), s)

    sql = parameterize_string(raw_sql)

    print('parameter sql', sql)

    def create_parameter(i, v):
        value_type = get_data_type(v)
        value = {value_type: v}
        if value_type == 'arrayValue':
            if v:
                element_type = '{}s'.format(get_data_type(v[0]))
            else:
                element_type = 'stringValues'
            value[value_type] = {element_type: v}
        elif v is 'null':
            value = {'isNull': True}

        return {'name': 'name{}'.format(i), 'value': value}

    parameters = [create_parameter(i, v) for i, v in enumerate(data)]
    return sql, parameters


def remove_special(s):
    return re.sub(r'[^a-zA-Z0-9]+', ' ', s)


def quote_escape(x):
    return parse.quote(str(x), '| @:')



def try_parse_date(s):
    try:
        return datetime.strptime(s, "%Y-%m-%dT%H:%M:%S.%fZ")
    except:
        try:
            s = s.split('.')[0] # Strip ms if present.
            return datetime.strptime(s, '%Y-%m-%d %H:%M:%S')
        except:
            return None


def stringify_for_data_api_query(x, manual_escape=False):
    type_x = type(x)

    if type_x in (bool, list, tuple, bytes, float, int):
        return x
    elif type_x is dict:
        return json.dumps(x)
    elif x is None:
        return 'null'

    if manual_escape:
        v = try_parse_date(x)
        if v:
            return str(v)
        return quote_escape(x)

    # Default to string.
    return str(x)


def make_list(items):
    return '({})'.format(','.join(map(lambda x: "'{}'".format(stringify_for_data_api_query(x, True)), items)))


# Convert comma separated list of items as string to escaped sql string list format
# ex: "'1', '2'" => ('1', '2')
def format_string_list(s):
    return make_list(re.sub("['\" ]", "", s).split(','))


# select unique items filtering nulls
def unique_on_key(elements, key):
    return list({element.get(key): element for element in elements if element and key in element}.values())


# https://stackoverflow.com/questions/40230546/python-string-percent-sign-escape/40230658
def escape_percent(value):
    return value.replace('%', '%%') if isinstance(value, str) else value


# https://realpython.com/prevent-python-sql-injection/
def create_update_clause(obj):
    clause = []
    for k, v in obj.items():
        k = quote_escape(k)
        if v is None:
            clause.append("{}=null".format(k))
        else:
            clause.append("{}='{}'".format(k, stringify_for_data_api_query(escape_percent(v), True)))
    if not clause:
        return ''

    return ', '.join(clause)
